package eu.transkribus.swt_gui.dialogs;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.fieldassist.AutoCompleteField;
import org.eclipse.jface.fieldassist.ComboContentAdapter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import eu.transkribus.core.model.beans.TrpPage;
import eu.transkribus.core.model.beans.TrpTranscriptMetadata;
import eu.transkribus.core.model.beans.enums.EditStatus;
import eu.transkribus.core.util.EnumUtils;
import eu.transkribus.swt.util.DialogUtil;
import eu.transkribus.swt.util.SWTUtil;
import eu.transkribus.swt_gui.util.CurrentTranscriptOrCurrentDocPagesSelector;

public class SetOlderVersionsAsCurrentDialog extends Dialog {
	private static final Logger logger = LoggerFactory.getLogger(SetOlderVersionsAsCurrentDialog.class);
	
//	CurrentDocPagesSelector pagesSelector;
	CurrentTranscriptOrCurrentDocPagesSelector pagesSelector;
//	LabeledText threshPercText;
	Button dryRunBtn;
	
//	ComboViewer threshPercComboViewer;
	Combo threshPercCombo;
	
	String status;
	String toolname;
	
	Set<Integer> pageIndices=null;
	boolean dryRun=true;

	public SetOlderVersionsAsCurrentDialog(Shell parentShell, String status, String toolname) {
		super(parentShell);
		this.status = status;
		this.toolname = toolname;
	}
	
	@Override
	protected Point getInitialSize() {
		return new Point(400, getShell().computeSize(SWT.DEFAULT, SWT.DEFAULT).y);
	}
	
	@Override
	protected void setShellStyle(int newShellStyle) {           
	    super.setShellStyle(SWT.CLOSE | SWT.MODELESS| SWT.BORDER | SWT.TITLE);
	    setBlockOnOpen(false);
	}
	
	@Override
	protected void configureShell(Shell newShell) {
		super.configureShell(newShell);
		newShell.setText("Set older versions as current");
	}
	
	@Override
	protected boolean isResizable() {
	    return true;
	}
	
	@Override
	protected Control createDialogArea(Composite parent) {
		Composite cont = (Composite) super.createDialogArea(parent);
		cont.setLayout(new GridLayout(1, false));
		
		pagesSelector = new CurrentTranscriptOrCurrentDocPagesSelector(cont, SWT.NONE, true,true);
		pagesSelector.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		
		Composite threshContainer = new Composite(cont, 0);
		threshContainer.setLayout(SWTUtil.createGridLayout(2, false, 0, 0));
		threshContainer.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		
		Label threshLabel = new Label(threshContainer, 0);
		threshLabel.setText("Select status or toolname: ");
		threshPercCombo = new Combo(threshContainer, SWT.DROP_DOWN);
		threshPercCombo.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		AutoCompleteField threshAutocomplete = new AutoCompleteField(threshPercCombo, new ComboContentAdapter(), new String[] {});
		String[] items = {status, toolname};
		
		threshPercCombo.setItems(items);
		threshPercCombo.select(1);
		threshPercCombo.setToolTipText("The version with exactly this status or toolname will be stored as current - this will be done for all selected pages");
		
//		threshPercText = new LabeledText(cont, "Percent of image area: ");
//		threshPercText.getTextField().setText(""+threshPerc);
//		threshPercText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
//		threshPercText.setToolTipText("All text-regions smaller than this percent of the image size are getting removed");
		
		dryRunBtn = new Button(cont, SWT.CHECK);
		dryRunBtn.setText("Dry run");
		dryRunBtn.setToolTipText("Does not save versions. This shows in how many pages versions with this status or toolname were found.");
		dryRunBtn.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		dryRunBtn.setSelection(dryRun);
		
		return cont;
	}
	
	@Override
	protected void okPressed() {
		try {
			pageIndices = pagesSelector.getSelectedPageIndices();
		} catch (IOException e) {
			pageIndices = null;
			DialogUtil.showErrorMessageBox(getShell(), "Invalid value", "Could not parse selected pages");
			logger.error("Could not parse page indices: "+e.getMessage(), e);
			return;
		}
		
		try {
//			threshPerc = Double.valueOf(threshPercText.getText());1remoptremote
		} catch (Exception e) {
			DialogUtil.showErrorMessageBox(getShell(), "Invalid value", "Could not parse threshold (must be a valid floating point number)");
			logger.error(e.getMessage());
			return;
		}
		
		dryRun = dryRunBtn.getSelection();
		
		logger.debug("pageIndices = "+pageIndices);
		logger.debug("dryRun = "+dryRun);
		
		super.okPressed();
	}

	public Set<Integer> getPageIndices() {
		return pageIndices;
	}


	public boolean isDryRun() {
		return dryRun;
	}

}
